--[[
This sample script shows how to use the cda module (and its submodules) to
generate a CDA document.  Data that is hardcoded in this project would instead
be generate from another data source (a database, for example).
--]]

require('cda')

local function trace(A,B,C) end

header = {}
body = {}
body.allergies = {}
body.problems = {}
body.medications = {}
body.results = {}
body.procedures = {}

local ID_ROOT = '2.16.840.1.113883.3.72'
local SET_ID_ROOT = '2.16.840.1.113883.19.7'

-- Override constant values
cda.XML_HEADER = [[<?xml version="1.0" encoding="UTF-8"?>
<!-- <?xml-stylesheet type="text/xsl" href="ccd.xsl"?> --> 
<!-- Adapted by Drummond Group Inc. (March 2011) from  the NIST original file "MU_Rev3_HITSP_C32C83_4Sections_RobustEntries_NoErrors.xml by Len Gallagher (NIST)"
       direct link: http://xreg2.nist.gov/cda-validation/downloads/muValidation20101013.tar.gz
        web site: http://xreg2.nist.gov/cda-validation/mu.html -->
]]

cda.SCHEMA_LOCATION = 'http://xreg2.nist.gov:8080/hitspValidation/schema/cdar2c32/infrastructure/cda/C32_CDA.xsd'


-- The main function is the first function called from Iguana.
-- The Data argument will contain the message to be processed.
function main()
   local Success = true
   local CDA
   
   local PatientId = '989285998'
   if iguana.isTest() then
      Cda = GenerateCDA(PatientId)
   else   
      Success, Cda = pcall(GenerateCDA, PatientId)
   end
   
   -- Easy way to validate: go to
   -- http://hit-testing.nist.gov/cda-validation/validation.html
   -- and select the file written out in the line below, then
   -- "Validate". Now, any time you want to see new validation
   -- results after running your script, simply reload the results
   -- page in your browser.  The browser should ask you if you want to
   -- resend the request, and when you say "yes", the new contents of
   -- the file will be validated.
   writeToFile('cda_2_out.xml', Cda)
   
   if not Success then
      net.http.respond{body=Cda, entity_type='text/plain'}
   else
      net.http.respond{body=Cda, entity_type='text/xml'}
   end
end

function GenerateCDA(PatientId)
   local Doc = cda.new()
   cda.setAttr(Doc, 'xmlns', 'sdtc', 'urn:hl7-org:sdtc')
   
   -- Header
   header.document(Doc)
   header.patient(Doc, PatientId)
   header.author(Doc)
   header.custodian(Doc)
   header.participants(Doc)
   header.documentOf(Doc)
   
   -- Body
   local Body = cda.createBody(Doc)
   
   -- Allergies
   local AAR, AT = body.allergies.new(Body)
   body.allergies.penicillin(AAR, AT)
   body.allergies.aspirin(AAR, AT)
   body.allergies.codeine(AAR, AT)
   
   -- Problems
   local Prob, ProbT = body.problems.new(Body)
   body.problems.heartFailure(Prob, ProbT)
   body.problems.myocardialInfarction(Prob, ProbT)
   
   -- Medications
   local M, MT = body.medications.new(Body)
   body.medications.captopril(M, MT)
   body.medications.spironolactone(M, MT)
   body.medications.digoxin(M, MT)
   body.medications.potassiumChloride(M, MT)
   
   -- Results
   local R, RT = body.results.new(Body)
   -- This one's done a little differently.
   -- This is more similar to how one would
   -- actually populate a section with contents
   -- from a database.
   body.results.observations(R, RT)
   
   -- Procedures
   local Proc, ProcT = body.procedures.new(Body)
   body.procedures.angioplasty(Proc, ProcT)
   body.procedures.catheterization(Proc, ProcT)
   
   return cda.serialize(Doc)
end

function header.document(Doc)
   local CD = Doc.ClinicalDocument
   cda.document.addTemplateId(CD, '2.16.840.1.113883.3.27.1776', 'CDA/R2')
   cda.document.addTemplateId(CD, '2.16.840.1.113883.10.20.3', 'HL7/CDT Header')
   cda.document.addTemplateId(CD, '1.3.6.1.4.1.19376.1.5.3.1.1.1', 'IHE/PCC')
   cda.document.addTemplateId(CD, '2.16.840.1.113883.3.88.11.32.1', 'HITSP/C32')
   cda.document.setId(CD, ID_ROOT, '170.306-(f)_DataSet-1', 'Drummond Group Inc')
   cda.document.setCode(CD, '34133-9', 'Summarization of episode node')
   cda.document.setTitle(CD, 'Continuity of Care Document - NIST 170.306(f) data set #1')
   cda.document.setEffectiveTime(CD, '2010-10-26 13:09:45', 6)
   cda.document.setConfidentialityCode(CD, 'N', 'Normal')
end

function header.patient(Doc, PatientId)
   local PR = cda.header.addPatientRole(
      Doc, '1.3.6.4.1.4.1.2835.3', PatientId, 'Good Health Clinic')
   cda.header.addAddr(PR, {street='353 Wine Street', city='Flint',
         state='MI', postalCode='48503', country='USA'}, 'HP')
   cda.header.addAddr(PR, {street='1000 Gulf of Mexico Drive #103',
         city='Longboat Key', state='FL', postalCode='34237', country='USA'}, 'HV')
   cda.header.addTelecom(PR, 'tel:+1-810-673-8378', 'HP')
   cda.header.addTelecom(PR, 'tel:+1-941-555-5555', 'HV')
   cda.header.addTelecom(PR, 'mailto:amarietoliver@hostname.com')
   cda.header.patient.addName(PR,  {family='Toliver', given='Ann'}, 'L')
   
   -- If we really need qualifiers here, we can use the basic XML API to do them.
   local Marie = cda.header.patient.addName(PR,  {family='Toliver', given='Marie'}, 'P')
   Marie.given:setAttr('qualifier', 'CL')
   
   -- If we really need multiple given names, we can use the basic XML API to do it.
   -- The second table parameter can simply be an empty table, and we can fill in
   -- the various name components ourselves.
   local AnnMarie = cda.header.patient.addName(PR, {}, 'P')
   AnnMarie:append(xml.ELEMENT, 'given'):setAttr('qualifier', 'BR'):setText('Ann')
   AnnMarie:append(xml.ELEMENT, 'given'):setAttr('qualifier', 'BR'):setText('Marie')
   AnnMarie:append(xml.ELEMENT, 'family'):setAttr('qualifier', 'BR'):setText('Toliver')
   
   cda.header.patient.setGender(PR, 'F', 'Female')
   cda.header.patient.setBirthtime(PR, '16-Jul-1950')
   cda.header.patient.setMaritalStatus(PR, 'W', 'Widowed')
   cda.header.patient.setRace(PR, '2058-6', 'African American')
   cda.header.patient.setEthnicGroup(PR, '2178-2', 'Latin American')
   
   local G = cda.header.patient.addGuardian(PR, 'GRMTH', 'Grandmother')
   cda.header.addAddr(G, {street='353 Wine Street', city='Flint', state='MI',
         postalCode='48503', country='USA'}, 'HP')
   cda.header.addTelecom(G, 'tel:+1-810-673-8378', 'HP')
   cda.header.patient.guardian.addName(G, {given='Elizabeth', family='Smith'}, 'L')
   
   cda.header.patient.addLanguageCommunication(PR, 'en-US', true)
   cda.header.patient.addLanguageCommunication(PR, 'es-US', false, 'RWR', 'Recieve Written')
end

function header.author(Doc)
   local A = cda.header.addAuthor(Doc)
   cda.header.author.setTime(A, '2010-10-26 13:09:45', 6)
   local AA = cda.header.author.addAssignedAuthor(A, '89753db2-4a15-4db8-a82e-d385a89e5617')
   
   cda.header.addAddr(AA, {street={'13359 North Hwy 183', 'Ste B-406-238'},
         city='Austin', state='TX', postalCode='78750', country='USA'})
   cda.header.addTelecom(AA, 'mailto:info2@drummondgroup.com ')
   cda.header.addTelecom(AA, 'tel:+1(817)-294-7339', 'WP')
   cda.header.assigned.addName(AA, {given='Test', family='Proctor'})
   
   local RO = cda.header.assigned.addRepresentedOrganization(AA, nil, nil, 'Drummond Group Inc')
   cda.header.addTelecom(RO, 'mailto:info2@drummondgroup.com ')
   cda.header.addTelecom(RO, 'tel:+1(817)-294-7339', 'WP')
   cda.header.addAddr(RO, {street={'13359 North Hwy 183', 'Ste B-406-238'},
         city='Austin', state='TX', postalCode='78750', country='USA'})
end

function header.custodian(Doc)
   local C = cda.header.addCustodian(Doc)
   local AC, RCO = cda.header.custodian.addAssignedCustodian(
      C, 'Good Health Clinic', '1.3.6.4.1.4.1.2835.3', '1')
   cda.header.addTelecom(RCO, 'tel:(999)555-1212', 'WP')
   cda.header.addAddr(RCO, {street='21 North Ave', city='Burlington',
         state='MA', postalCode='01803', country='USA'})
end

function header.participants(Doc)
   local P1 = cda.header.addParticipant(Doc, 'IND')
   cda.participant.setTime(P1, '2010', 1)
   local AE = cda.participant.addAssociatedEntity(P1, 'NOK', 'MTH', 'Mother')
   cda.header.addAddr(AE, {street='353 Wine Street', city='Flint', state='MI',
         postalCode='48503', country='USA'}, 'HP')
   cda.header.addTelecom(AE, 'tel:+1-810-673-8378', 'HP')
   cda.header.associated.addName(AE, 'Marie Williams')
   
   local P2 = cda.header.addParticipant(Doc, 'IND')
   cda.participant.setTime(P2, '2010', 1)
   local AE2 = cda.participant.addAssociatedEntity(P2, 'ECON', 'MTH', 'Mother')
   cda.header.addAddr(AE2, {street='353 Wine Street', city='Flint', state='MI',
         postalCode='48503', country='USA'}, 'HP')
   cda.header.addTelecom(AE2, 'tel:+1-810-673-8378', 'HP')
   cda.header.associated.addName(AE2, 'Marie Williams')
end

function header.documentOf(Doc)
   local SE = cda.header.addServiceEvent(Doc, 'DOC', 'PCPR')
   local ET = cda.header.serviceevent.addEffectiveTime(SE)
   cda.util.date.setRange(ET, {value='2002-06-01'}, {value='2010-10-26'})
   
   -- First "performer"
   do
      local P = cda.header.serviceevent.addPerformer(SE, 'PRF', 'PP', 'Primary Care Provider')
      local T = cda.header.serviceevent.performer.addTime(P)
      cda.util.date.setRange(T, {value='1950-07-16'}, {value='2011-04-27'})
      local AE = cda.header.serviceevent.performer.addAssignedEntity(
         P, '2.16.840.1.113883.4.6', 'Unknown', 'NPI')
      cda.header.assigned.addId(
         AE, '2.16.840.1.113883.3.72.5.2', 'PseudoMD-1', 'NIST Healthcare Testing Laboratory')
      cda.header.assigned.addProviderCode(AE, '200000000X', 'Allopathic and Osteopathic Physicians')
      cda.header.addAddr(AE, {}) -- empty
      cda.header.addTelecom(AE) -- empty
      cda.header.assigned.addName(AE, {prefix='Dr.', given='James', family='Smith'})
      
      local RO = cda.header.assigned.addRepresentedOrganization(
         AE, '2.16.840.1.113883.3.72.5', nil, 'County Hospital')
      cda.header.addTelecom(RO, 'tel:+1-810-555-6790', 'WP')
      cda.header.addAddr(RO, {street='6258 Main St.', city='Flint',
            state='MI', postalCode='48503', country='USA'}, 'WP')
   end
   
   -- Second "performer"
   do
      local P = cda.header.serviceevent.addPerformer(SE, 'PRF', 'PP', 'Primary Care Provider')
      local T = cda.header.serviceevent.performer.addTime(P)
      cda.util.date.setRange(T, {value='2002-07-16'}, {value='2007-09-15'})
      local AE = cda.header.serviceevent.performer.addAssignedEntity(
         P, '2.16.840.1.113883.4.6', 'Unknown', 'NPI')
      cda.header.assigned.addId(
         AE, '2.16.840.1.113883.3.72.5.2', 'PseudoMD-2', 'Drummond Group Inc')
      cda.header.assigned.addProviderCode(AE, '200000000X', 'Allopathic and Osteopathic Physicians')
      cda.header.addAddr(AE, {street='9560 Main St.', city='Flint',
            state='MI', postalCode='48503', country='USA'}, 'WP')
      cda.header.addTelecom(AE, 'tel:+1-810-555-3251', 'WP')
      cda.header.assigned.addName(AE, {prefix='Dr.', given='Barry', family='Stein'})
      
      local RO = cda.header.assigned.addRepresentedOrganization(
         AE, '2.16.840.1.113883.3.72.5', nil, 'Memorial Hospital')
      cda.header.addTelecom(RO, 'tel:+1-810-555-3269', 'WP')
      cda.header.addAddr(RO, {street='9588 Main St.', city='Flint',
            state='MI', postalCode='48503', country='USA'}, 'WP')
   end
end

local ALLERGIES_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.102', assigningAuthorityName='HITSP/C83'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.3.13', assigningAuthorityName='IHE PCC'},
   {root='2.16.840.1.113883.10.20.1.2', assigningAuthorityName='HL7 CCD'}
}

local ALLERGIES_TABLE = [[
<table border="1" width="100%">
   <thead>
      <tr>
         <th>Type</th>
         <th>Substance</th>
         <th>Reaction</th>
         <th>Status</th>
      </tr>
   </thead>
   <tbody>
   </tbody>
</table>
]]

function body.allergies.new(Body)
   local S = cda.section.new(Body, '48765-2', 'Allergies and Adverse Reactions',
      ALLERGIES_TEMPLATE_IDS, 'Allergies')
   
   local T = cda.section.addText(S)
   T:setInner(ALLERGIES_TABLE)
   
   return S, T.table.tbody
end

local ALLERGIES_TABLE_ROW = [[
<td ID="#">#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td ID="#">#</td>
]]

local function addAllergySummary(T, Index, Algtype, Algsub, Algreact, Algstatus)
   local TR = T:append(xml.ELEMENT, 'tr')
   TR:setAttr('ID', 'ALGSUMMARY_'..Index)
   TR:setInner(ALLERGIES_TABLE_ROW)
   TR[2].ID = 'ALGTYPE_'..Index
   TR[2][2] = Algtype
   TR[3].ID = 'ALGSUB_'..Index
   TR[3][2] = Algsub
   TR[4].ID = 'ALGREACT_'..Index
   TR[4][2] = Algreact
   TR[5].ID = 'ALGSTATUS_'..Index
   TR[5][2] = Algstatus
end

local ALLERGY_ACT_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.6', assigningAuthorityName='HITSP C83'},
   {root='2.16.840.1.113883.10.20.1.27', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.1', assigningAuthorityName='IHE PCC'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.3', assigningAuthorityName='IHE PCC'}
}

local ALLERGY_OBSERVATION_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.18', assigningAuthorityName='CCD'},
   {root='2.16.840.1.113883.10.20.1.28', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5', assigningAuthorityName='IHE PCC'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.6', assigningAuthorityName='IHE PCC'}
}

function body.allergies.penicillin(S, T)
   addAllergySummary(T, 1, 'Drug Allergy', 'Penicillin', 'Hives', 'Active')
   
   local A = cda.section.addActEntry(S, 'ACT', 'EVN', 'DRIV')
   cda.util.addTemplateIds(A, ALLERGY_ACT_TEMPLATE_IDS)
   cda.entry.setIdRoot(A, '36e3e930-7b14-11db-9fe1-0800200c9a66')
   cda.entry.addNullFlavor(A, 'NA')
   cda.entry.setStatusCode(A, 'completed')
   local AET = cda.entry.setEffectiveTime(A)
   cda.util.date.setRange(AET, {nullFlavor='UNK'}, {nullFlavor='UNK'})
   
   local O = cda.observation.addRelatedObservation(
      A, 'SUBJ', 'OBS', 'EVN', {inversionInd=false})
   cda.util.addTemplateIds(O, ALLERGY_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, '4adc1020-7b14-11db-9fe1-0800200c9a66')
   cda.observation.addCode(O, '416098002', cda.cs.SNOMED_CT, 'drug allergy')
   cda.observation.addTextReference(O, '#ALGSUMMARY_1')
   cda.entry.setStatusCode(O, 'completed')
   local OET = cda.observation.setEffectiveTime(O)
   cda.util.date.setRange(OET, {nullFlavor='UNK'})
   local V = cda.observation.addValueWithCode(
      O, '70618', cda.cs.RxNorm, 'Penicillin', 'CD')
   cda.entry.code.addOriginalTextReference(V, '#ALGSUB_1')
   
   local PR = cda.observation.addParticipantRole(O, 'CSM', 'MANU')
   local PE = cda.participant.addPlayingEntity(
      PR, 'MMAT', '70618', cda.cs.RxNorm, 'Penicillin', '#ALGSUB_1')
   
   local RO = cda.observation.addRelatedObservation(
      O, 'MFST', 'OBS', 'EVN', {inversionInd=true})
   cda.util.addTemplateIds(RO, {{root='2.16.840.1.113883.10.20.1.54'}})
   cda.observation.addCode(RO, 'ASSERTION', cda.cs.IntoleranceType)
   cda.observation.setStatusCode(RO, 'completed')
   local RV = cda.observation.addValueWithCode(
      RO, '247472004', cda.cs.SNOMED_CT, 'Hives', 'CD')
   cda.observation.code.addOriginalTextReference(RV, '#ALGREACT_1')
   
   local SO = cda.observation.addRelatedObservation(O, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.39'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local SV = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(SV, '#ALGSTATUS_1')
end

function body.allergies.aspirin(S, T)
   addAllergySummary(T, 2, 'Drug Intolerance', 'Aspirin', 'Wheezing', 'Active')
   
   local A = cda.section.addActEntry(S, 'ACT', 'EVN', 'DRIV')
   cda.util.addTemplateIds(A, ALLERGY_ACT_TEMPLATE_IDS)
   cda.entry.setIdRoot(A, 'eb936010-7b17-11db-9fe1-0800200c9a66')
   cda.entry.addNullFlavor(A, 'NA')
   cda.entry.setStatusCode(A, 'active')
   local AET = cda.entry.setEffectiveTime(A)
   cda.util.date.setRange(AET, {nullFlavor='UNK'})
   
   local O = cda.observation.addRelatedObservation(
      A, 'SUBJ', 'OBS', 'EVN', {inversionInd=false})
   cda.util.addTemplateIds(O, ALLERGY_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, 'eb936011-7b17-11db-9fe1-0800200c9a66')
   cda.observation.addCode(
      O, '419511003', cda.cs.SNOMED_CT, 'prepensity to adverse reactions to drug')
   cda.observation.addTextReference(O, '#ALGSUMMARY_2')
   cda.observation.setStatusCode(O, 'completed')
   local OET = cda.entry.setEffectiveTime(O)
   cda.util.date.setRange(OET, {nullFlavor='UNK'})
   local V = cda.observation.addValueWithCode(
      O, '1191', cda.cs.RxNorm, 'Aspirin', 'CD')
   cda.entry.code.addOriginalTextReference(V, '#ALGSUB_2')
   
   local PR = cda.observation.addParticipantRole(O, 'CSM', 'MANU')
   local PE = cda.participant.addPlayingEntity(
      PR, 'MMAT', '1191', cda.cs.RxNorm, 'Aspirin', '#ALGSUB_2')
   
   local RO = cda.observation.addRelatedObservation(
      O, 'MFST', 'OBS', 'EVN', {inversionInd=true})
   cda.util.addTemplateIds(RO, {{root='2.16.840.1.113883.10.20.1.54'}})
   cda.observation.addCode(RO, 'ASSERTION', cda.cs.IntoleranceType)
   cda.observation.setStatusCode(RO, 'completed')
   local RV = cda.observation.addValueWithCode(
      RO, '272040008', cda.cs.SNOMED_CT, 'Wheezing', 'CD')
   cda.observation.code.addOriginalTextReference(RV, '#ALGREACT_2')
   
   local SO = cda.observation.addRelatedObservation(O, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.39'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local SV = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(SV, '#ALGSTATUS_2')
end

function body.allergies.codeine(S, T)
   addAllergySummary(T, 3, 'Drug Intolerance', 'Codeine', 'Nausea', 'Active')
   
   local A = cda.section.addActEntry(S, 'ACT', 'EVN', 'DRIV')
   cda.util.addTemplateIds(A, ALLERGY_ACT_TEMPLATE_IDS)
   cda.entry.setIdRoot(A, 'c3df3b61-7b18-11db-9fe1-0800200c9a66')
   cda.entry.addNullFlavor(A, 'NA')
   cda.entry.setStatusCode(A, 'active')
   local AET = cda.entry.setEffectiveTime(A)
   cda.util.date.setRange(AET, {nullFlavor='UNK'})
   
   local O = cda.observation.addRelatedObservation(
      A, 'SUBJ', 'OBS', 'EVN', {inversionInd=false})
   cda.util.addTemplateIds(O, ALLERGY_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, 'c3df3b60-7b18-11db-9fe1-0800200c9a66')
   cda.observation.addCode(
      O, '59037007', cda.cs.SNOMED_CT, 'drug intolerance')
   cda.observation.addTextReference(O, '#ALGSUMMARY_3')
   cda.observation.setStatusCode(O, 'completed')
   local OET = cda.entry.setEffectiveTime(O)
   cda.util.date.setRange(OET,
      {value='2005-12-01', precision=2}, {value='2006-01-01', precision=2})
   local V = cda.observation.addValueWithCode(
      O, '2670', cda.cs.RxNorm, 'Codeine', 'CD')
   cda.entry.code.addOriginalTextReference(V, '#ALGSUB_3')
   
   local PR = cda.observation.addParticipantRole(O, 'CSM', 'MANU')
   local PE = cda.participant.addPlayingEntity(
      PR, 'MMAT', '2670', cda.cs.RxNorm, 'Codeine', '#ALGSUB_3')
   
   local RO = cda.observation.addRelatedObservation(
      O, 'MFST', 'OBS', 'EVN', {inversionInd=true})
   cda.util.addTemplateIds(RO,
      {{root='2.16.840.1.113883.10.20.1.54', assigningAuthorityName='CCD'}})
   cda.observation.addCode(RO, 'ASSERTION', cda.cs.IntoleranceType)
   cda.observation.addText(RO) -- empty
   cda.observation.setStatusCode(RO, 'completed')
   local RV = cda.observation.addValueWithCode(
      RO, '73879007', cda.cs.SNOMED_CT, 'Nausea', 'CD')
   cda.observation.code.addOriginalTextReference(RV, '#ALGREACT_3')
   
   local RSO = cda.observation.addRelatedObservation(RO, 'SUBJ', 'OBS', 'EVN')
   cda.util.addTemplateIds(RSO,
      {{root='2.16.840.1.113883.10.20.1.55', assigningAuthorityName='CCD'}})
   cda.observation.addCode(RSO, 'SEV', cda.cs.ActCode, 'Severity')
   cda.observation.addText(RSO, 'Required by HITSP C-83')
   cda.observation.setStatusCode(RSO, 'completed')
   cda.observation.addValueWithCode(RSO, '6736007', cda.cs.SNOMED, 'moderate', 'CE')
   
   local SO = cda.observation.addRelatedObservation(O, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.39'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local SV = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(SV, '#ALGSTATUS_3')
end

local PROBLEMS_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.103', assigningAuthorityName='HITSP/C83'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.3.6', assigningAuthorityName='IHE PCC'},
   {root='2.16.840.1.113883.10.20.1.11', assigningAuthorityName='HL7 CCD'}
}

local PROBLEMS_TABLE = [[
<table border="1" width="100%">
   <thead>
      <tr>
         <th>ICD-9-CM Code</th>
         <th>Patient Problem</th>
         <th>Status</th>
         <th>Date Diagnosed</th>
      </tr>
   </thead>
   <tbody>
   </tbody>
</table>
]]

function body.problems.new(Body)
   local S = cda.section.new(Body, '11450-4', 'Problems',
      PROBLEMS_TEMPLATE_IDS, 'Problem List')
   
   local T = cda.section.addText(S)
   T:setInner(PROBLEMS_TABLE)
   
   return S, T.table.tbody
end

local PROBLEMS_TABLE_ROW = [[
<td>#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td>#</td>
]]

local function addProblemSummary(T, Index, Code, Kind, Status, DateDiagnosed)
   local TR = T:append(xml.ELEMENT, 'tr')
   TR:setAttr('ID', 'PROBSUMMARY_'..Index)
   TR:setInner(PROBLEMS_TABLE_ROW)
   TR[2][1] = Code
   TR[3].ID = 'PROBKIND_'..Index
   TR[3][2] = Kind
   TR[4].ID = 'PROBSTATUS_'..Index
   TR[4][2] = Status
   TR[5][1] = DateDiagnosed
end

local PROBLEM_ACT_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.7', assigningAuthorityName='HITSP C83'},
   {root='2.16.840.1.113883.10.20.1.27', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.1', assigningAuthorityName='IHE PCC'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.2', assigningAuthorityName='IHE PCC'}
}

local PROBLEM_OBSERVATION_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.28', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.5', assigningAuthorityName='IHE PCC'}
}

function body.problems.heartFailure(S, T)
   addProblemSummary(T, 1, '428.0', 'Congestive Heart Failure', 'Active', '2/22/2010')
   
   local A = cda.section.addActEntry(S, 'ACT', 'EVN', 'DRIV')
   cda.util.addTemplateIds(A, PROBLEM_ACT_TEMPLATE_IDS)
   cda.entry.setIdRoot(A, '6a2fa88d-4174-4909-aece-db44b60a3abb')
   cda.entry.addNullFlavor(A, 'NA')
   cda.entry.setStatusCode(A, 'completed')
   local AET = cda.entry.setEffectiveTime(A)
   cda.util.date.setRange(AET, {value='2010-02-22'}, {nullFlavor='UNK'})
   
   local P = cda.entry.addPerformer(A, 'PRF','2.16.840.1.113883.3.72.5.2', 'PseudoMD-2')
   cda.header.addAddr(P, {}) -- empty
   cda.header.addTelecom(P) -- empty
   
   local O = cda.entry.addRelatedObservation(A, 'SUBJ', 'OBS', 'EVN', {inversionInd=false})
   O:setAttr('negationInd', 'false') -- easy to just add this attribute if it's necessary
   cda.util.addTemplateIds(O, PROBLEM_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, 'd11275e7-67ae-11db-bd13-0800200c9a66')
   cda.observation.addCode(O, '64572001', cda.cs.SNOMED_CT, 'Condition')
   cda.observation.addTextReference(O, '#PROBSUMMARY_1')
   cda.observation.setStatusCode(O, 'completed')
   local OET = cda.entry.setEffectiveTime(O)
   cda.util.date.setRange(OET, {value='2010-02-22'}, {nullFlavor='UNK'})
   local V = cda.observation.addValue(O, '', 'CD')
   -- Example of how to tweak elements to meet your needs:
   V:remove('value') -- remove attrbitue
   V:setAttr('nullFlavor', 'UNK') -- add attribute
   local T = cda.observation.value.addTranslationWithDisplayName(
      V, '428.0', cda.cs.ICD9CM_Diagnosis, 'Congestive Heart Failure')
   cda.util.setXsiType(T, 'CD')
   
   local SO = cda.observation.addRelatedObservation(O, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.50'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local SV = cda.observation.addValueWithCode(SO, '55561003', cda.cs.SNOMED, 'Active', 'CE')
   cda.entry.code.addOriginalTextReference(SV, '#PROBSTATUS_1')
end

function body.problems.myocardialInfarction(S, T)
   addProblemSummary(T, 2, '410.90', 'Acute Myocardial Infarction', 'Resolved', '9/16/2007')
   
   local A = cda.section.addActEntry(S, 'ACT', 'EVN', 'DRIV')
   cda.util.addTemplateIds(A, PROBLEM_ACT_TEMPLATE_IDS)
   cda.entry.setIdRoot(A, 'ec8a6ff8-ed4b-4f7e-82c3-e98e58b45de7')
   cda.entry.addNullFlavor(A, 'NA')
   cda.entry.setStatusCode(A, 'completed')
   local AET = cda.entry.setEffectiveTime(A)
   cda.util.date.setRange(AET, {value='2007-09-16'}, {nullFlavor='UNK'})
   
   local P = cda.entry.addPerformer(A, 'PRF','2.16.840.1.113883.3.72.5.2', 'PseudoMD-1')
   cda.header.addAddr(P, {}) -- empty
   cda.header.addTelecom(P) -- empty
   
   local O = cda.entry.addRelatedObservation(A, 'SUBJ', 'OBS', 'EVN', {inversionInd=false})
   cda.util.addTemplateIds(O, PROBLEM_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, 'ab1791b0-5c71-11db-b0de-0800200c9a66')
   cda.observation.addCode(O, '64572001', cda.cs.SNOMED_CT, 'Condition')
   cda.observation.addTextReference(O, '#PROBSUMMARY_2')
   cda.observation.setStatusCode(O, 'completed')
   local OET = cda.entry.setEffectiveTime(O)
   cda.util.date.setRange(OET, {value='2007-09-16'}, {nullFlavor='UNK'})
   local V = cda.observation.addValue(O, '', 'CD')
   -- Example of how to tweak elements to meet your needs:
   V:remove('value') -- remove attrbitue
   V:setAttr('nullFlavor', 'UNK') -- add attribute
   local T = cda.observation.value.addTranslationWithDisplayName(
      V, '410.90', cda.cs.ICD9CM_Diagnosis, 'Acute Myocardial Infarction')
   cda.util.setXsiType(T, 'CD')
   
   local SO = cda.observation.addRelatedObservation(O, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.50'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local SV = cda.observation.addValueWithCode(SO, '413322009', cda.cs.SNOMED, 'Resolved', 'CE')
   cda.entry.code.addOriginalTextReference(SV, '#PROBSTATUS_2')
end

local MEDICATIONS_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.112', assigningAuthorityName='HITSP/C83'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.3.19', assigningAuthorityName='IHE PCC'},
   {root='2.16.840.1.113883.10.20.1.8', assigningAuthorityName='HL7 CCD'}
}

local MEDICATIONS_TABLE = [[
<table border="1" width="100%">
   <thead>
      <tr>
         <th>RxNorm Code</th>
         <th>Generic Name</th>
         <th>Brand Name</th>
         <th>Dose</th>
         <th>Form</th>
         <th>Route</th>
         <th>Frequency</th>
         <th>Sig Text</th>
         <th>Date Started</th>
         <th>Status</th>
      </tr>
   </thead>
   <tbody>
   </tbody>
</table>
]]

function body.medications.new(Body)
   local S = cda.section.new(Body, '10160-0', 'Medications',
      MEDICATIONS_TEMPLATE_IDS, 'History of medication use')
   
   local T = cda.section.addText(S)
   T:setInner(MEDICATIONS_TABLE)
   
   return S, T.table.tbody
end

local MEDICATIONS_TABLE_ROW_TEMPLATE = [[
<td>#</td>
<td ID="#">#</td>
<td>#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td ID="#">#</td>
<td>#</td>
<td ID="#">#</td>
]]

local function addMedicationSummary
(
   T, Index, Code, GenericName, BrandName, Dose, Form, Route,
   Frequency, SigText, DateStarted, Status
)
   local TR = T:append(xml.ELEMENT, 'tr')
   TR:setAttr('ID', 'MEDSUMMARY_'..Index)
   TR:setInner(MEDICATIONS_TABLE_ROW_TEMPLATE)
   TR[2][1] = Code
   TR[3].ID = 'MEDNAME_'..Index
   TR[3][2] = GenericName
   TR[4][1] = BrandName
   TR[5].ID = 'MEDDOSE_'..Index
   TR[5][2] = Dose
   TR[6].ID = 'MEDFORM_'..Index
   TR[6][2] = Form
   TR[7].ID = 'MEDROUTE_'..Index
   TR[7][2] = Route
   TR[8].ID = 'MEDFREQUENCY_'..Index
   TR[8][2] = Frequency
   TR[9].ID = 'SIGTXT_'..Index
   TR[9][2] = SigText
   TR[10][1] = DateStarted
   TR[11].ID = 'MEDSTATUS_'..Index
   TR[11][2] = Status
end

local MEDICATION_SA_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.8', assigningAuthorityName='HITSP C83'},
   {root='2.16.840.1.113883.10.20.1.24', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.7', assigningAuthorityName='IHE PCC'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.7.1', assigningAuthorityName='IHE PCC'}
}

local MANUFACTURED_PRODUCT_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.8.2', assigningAuthorityName='HITSP C83'},
   {root='2.16.840.1.113883.10.20.1.53', assigningAuthorityName='CCD'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.7.2', assigningAuthorityName='IHE PCC'}
}

function body.medications.captopril(S, T)
   addMedicationSummary(T, 1, '201372', 'Captopril 25 MG Oral Tablet',
      'Captopril 25 MG Oral Tablet [Capoten]', '1 tablet', 'tablet',
      'oral', 'TID', '25 mg PO TID (1 tablet by mouth 3 times per day)',
      '02/25/2010', 'Active')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN', 'DRIV')
   cda.util.addTemplateIds(SA, MEDICATION_SA_TEMPLATE_IDS)
   cda.substanceadmin.setIdRoot(SA, 'cdbd5b05-6cde-11db-9fe1-0800200c9a66')
   cda.substanceadmin.addTextReference(SA, '#SIGTEXT_1')
   cda.substanceadmin.setStatusCode(SA, 'completed')
   local SAET1 = cda.entry.setEffectiveTime(SA)
   cda.util.setXsiType(SAET1, 'IVL_TS')
   cda.util.date.setRange(SAET1, {value='2010-02-25'}, {nullFlavor='UNK'})
   local SAET2 = cda.substanceadmin.addEffectiveTime(SA, false, 8, 'h')
   SAET2:setAttr('operator', 'A')
   local RC = cda.substanceadmin.setRouteCode(
      SA, 'C38288', cda.cs.NCI_FDA_RouteOfAdministration, 'ORAL')
   cda.substanceadmin.code.addOriginalTextReference(RC, '#MEDROUTE_1')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      RC, 'PO', cda.cs.RouteOfAdministration, 'Swallow, oral')
   cda.substanceadmin.setDoseQuantity(SA, 1, '{TABLET}')
   -- A good example of how to add any generic item with a code and
   -- codeSystem (and displayName).
   local AUC = cda.util.addCodeElement(SA, 'administrationUnitCode',
      'C42998', cda.cs.NCI_FDA_DosageForms, 'TABLET')
   cda.substanceadmin.code.addOriginalTextReference(AUC, '#MEDFORM_1')
   
   local MM = cda.substanceadmin.addManufacturedMaterial(
      SA, 'MANU', 'MMAT', 'KIND', MANUFACTURED_PRODUCT_TEMPLATE_IDS)
   local MMC = cda.substanceadmin.manufactured.addCode(
      MM, '317173', cda.cs.RxNorm, 'Captopril 25 MG Oral Tablet')
   cda.substanceadmin.code.addOriginalTextReference(MMC, '#MEDNAME_1')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      MMC, '201372', cda.cs.RxNorm, 'Captopril 25 MG Oral Tablet [Capoten]')
   cda.substanceadmin.manufactured.setName(MM, 'Capoten')
   
   local SO = cda.substanceadmin.addRelatedObservation(SA, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.47'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(SO, 'completed')
   local V = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#MEDSTATUS_1')
   
   local PDO = cda.substanceadmin.addRelatedObservation(SA, 'SUBJ', 'OBS', 'EVN')
   cda.util.addTemplateIds(PDO,
      {{root='2.16.840.1.113883.3.88.11.83.8.1', assigningAuthorityName='HITSP C83'}})
   cda.observation.addCode(PDO, '73639000', cda.cs.SNOMED_CT, 'Prescription Drug')
end

function body.medications.spironolactone(S, T)
   addMedicationSummary(T, 2, '200820', 'Spironolactone 25 MG Oral Tablet',
      'Spironolactone 25 MG Oral Tablet [Aldactone]', '1 tablet', 'tablet',
      'oral', 'QID', '25 mg PO QID (1 tablet by mouth 4 times per day)',
      '02/25/2010', 'Active')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN', 'DRIV')
   cda.util.addTemplateIds(SA, MEDICATION_SA_TEMPLATE_IDS)
   cda.substanceadmin.setIdRoot(SA, 'aa554db5-c90d-4542-8d8a-2e00d79c9969')
   cda.substanceadmin.addTextReference(SA, '#SIGTEXT_2')
   cda.substanceadmin.setStatusCode(SA, 'completed')
   local SAET1 = cda.entry.setEffectiveTime(SA)
   cda.util.setXsiType(SAET1, 'IVL_TS')
   cda.util.date.setRange(SAET1, {value='2010-02-25'}, {nullFlavor='UNK'})
   local SAET2 = cda.substanceadmin.addEffectiveTime(SA, false, 6, 'h')
   SAET2:setAttr('operator', 'A')
   local RC = cda.substanceadmin.setRouteCode(
      SA, 'C38288', cda.cs.NCI_FDA_RouteOfAdministration, 'ORAL')
   cda.substanceadmin.code.addOriginalTextReference(RC, '#MEDROUTE_2')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      RC, 'PO', cda.cs.RouteOfAdministration, 'Swallow, oral')
   cda.substanceadmin.setDoseQuantity(SA, 1, '{TABLET}')
   -- A good example of how to add any generic item with a code and
   -- codeSystem (and displayName).
   local AUC = cda.util.addCodeElement(SA, 'administrationUnitCode',
      'C42998', cda.cs.NCI_FDA_DosageForms, 'TABLET')
   cda.substanceadmin.code.addOriginalTextReference(AUC, '#MEDFORM_2')
   
   local MM = cda.substanceadmin.addManufacturedMaterial(
      SA, 'MANU', 'MMAT', 'KIND', MANUFACTURED_PRODUCT_TEMPLATE_IDS)
   local MMC = cda.substanceadmin.manufactured.addCode(
      MM, '313096', cda.cs.RxNorm, 'Spironolactone 25 MG Oral Tablet')
   cda.substanceadmin.code.addOriginalTextReference(MMC, '#MEDNAME_2')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      MMC, '200820', cda.cs.RxNorm, 'Spironolactone 25 MG Oral Tablet [Aldactone]')
   cda.substanceadmin.manufactured.setName(MM, 'Aldactone')
   
   local PDO = cda.substanceadmin.addRelatedObservation(SA, 'SUBJ', 'OBS', 'EVN')
   cda.util.addTemplateIds(PDO,
      {{root='2.16.840.1.113883.3.88.11.83.8.1', assigningAuthorityName='HITSP C83'}})
   cda.observation.addCode(PDO, '73639000', cda.cs.SNOMED_CT, 'Prescription Drug')
   
   local SO = cda.substanceadmin.addRelatedObservation(SA, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.47'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   local V = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#MEDSTATUS_2')
end

function body.medications.digoxin(S, T)
   addMedicationSummary(T, 3, '309888', 'Digoxin 0.125 MG Oral Tablet',
      'Digoxin 0.125 MG Oral Tablet [Lanoxin', '1 tablet', 'tablet',
      'oral', 'Q Day', '125 mcg PO Q Day (1 tablet by mouth daily)',
      '02/25/2010', 'Active')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN', 'DRIV')
   cda.util.addTemplateIds(SA, MEDICATION_SA_TEMPLATE_IDS)
   cda.substanceadmin.setIdRoot(SA, '60f74272-f4e1-4ed3-b5f4-2932e0346a91')
   cda.substanceadmin.addTextReference(SA, '#SIGTEXT_3')
   cda.substanceadmin.setStatusCode(SA, 'completed')
   local SAET1 = cda.entry.setEffectiveTime(SA)
   cda.util.setXsiType(SAET1, 'IVL_TS')
   cda.util.date.setRange(SAET1, {value='2010-02-25'}, {nullFlavor='UNK'})
   local SAET2 = cda.substanceadmin.addEffectiveTime(SA, false, 24, 'h')
   SAET2:setAttr('operator', 'A')
   local RC = cda.substanceadmin.setRouteCode(
      SA, 'C38288', cda.cs.NCI_FDA_RouteOfAdministration, 'ORAL')
   cda.substanceadmin.code.addOriginalTextReference(RC, '#MEDROUTE_3')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      RC, 'PO', cda.cs.RouteOfAdministration, 'Swallow, oral')
   cda.substanceadmin.setDoseQuantity(SA, 1, '{TABLET}')
   -- A good example of how to add any generic item with a code and
   -- codeSystem (and displayName).
   local AUC = cda.util.addCodeElement(SA, 'administrationUnitCode',
      'C42998', cda.cs.NCI_FDA_DosageForms, 'TABLET')
   cda.substanceadmin.code.addOriginalTextReference(AUC, '#MEDFORM_3')
   
   local MM = cda.substanceadmin.addManufacturedMaterial(
      SA, 'MANU', 'MMAT', 'KIND', MANUFACTURED_PRODUCT_TEMPLATE_IDS)
   local MMC = cda.substanceadmin.manufactured.addCode(
      MM, '197604', cda.cs.RxNorm, 'Digoxin 0.125 MG Oral Tablet')
   cda.substanceadmin.code.addOriginalTextReference(MMC, '#MEDNAME_3')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      MMC, '309888', cda.cs.RxNorm, 'Digoxin 0.125 MG Oral Tablet [Lanoxin]')
   cda.substanceadmin.manufactured.setName(MM, 'Lanoxin')
   
   local PDO = cda.substanceadmin.addRelatedObservation(SA, 'SUBJ', 'OBS', 'EVN')
   cda.util.addTemplateIds(PDO,
      {{root='2.16.840.1.113883.3.88.11.83.8.1', assigningAuthorityName='HITSP C83'}})
   cda.observation.addCode(PDO, '73639000', cda.cs.SNOMED_CT, 'Prescription Drug')
   
   local SO = cda.substanceadmin.addRelatedObservation(SA, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.47'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   local V = cda.observation.addValueWithCode(
      SO, '55561003', cda.cs.SNOMED_CT, 'Active', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#MEDSTATUS_3')
end

function body.medications.potassiumChloride(S, T)
   addMedicationSummary(T, 4, '628958', 'Potassium Chloride 10 MEQ Extended Release Tablet',
      'Potassium Chloride 10 MEQ Extended Release Tablet [Klor-Con]', '1 tablet', 'tablet',
      'oral', 'BID', '10 mEq PO BID (1 tablet by mouth twice daily)',
      '02/25/2010', 'No Longer Active')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN', 'DRIV')
   cda.util.addTemplateIds(SA, MEDICATION_SA_TEMPLATE_IDS)
   cda.substanceadmin.setIdRoot(SA, '4b30dfed-1544-4c82-8e9d-7093461a04fd')
   cda.substanceadmin.addTextReference(SA, '#SIGTEXT_4')
   cda.substanceadmin.setStatusCode(SA, 'completed')
   local SAET1 = cda.entry.setEffectiveTime(SA)
   cda.util.setXsiType(SAET1, 'IVL_TS')
   cda.util.date.setRange(SAET1, {value='2010-02-25'})
   local SAET2 = cda.substanceadmin.addEffectiveTime(SA, false, 12, 'h')
   SAET2:setAttr('operator', 'A')
   local RC = cda.substanceadmin.setRouteCode(
      SA, 'C38288', cda.cs.NCI_FDA_RouteOfAdministration, 'ORAL')
   cda.substanceadmin.code.addOriginalTextReference(RC, '#MEDROUTE_4')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      RC, 'PO', cda.cs.RouteOfAdministration, 'Swallow, oral')
   cda.substanceadmin.setDoseQuantity(SA, 1, '{TABLET}')
   -- A good example of how to add any generic item with a code and
   -- codeSystem (and displayName).
   local AUC = cda.util.addCodeElement(SA, 'administrationUnitCode',
      'C42998', cda.cs.NCI_FDA_DosageForms, 'TABLET')
   cda.substanceadmin.code.addOriginalTextReference(AUC, '#MEDFORM_4')
   
   local MM = cda.substanceadmin.addManufacturedMaterial(
      SA, 'MANU', 'MMAT', 'KIND', MANUFACTURED_PRODUCT_TEMPLATE_IDS)
   local MMC = cda.substanceadmin.manufactured.addCode(
      MM, '628953', cda.cs.RxNorm,
      'Potassium Chloride 10 MEQ Extended Release Tablet [Klor-Con]')
   cda.substanceadmin.code.addOriginalTextReference(MMC, '#MEDNAME_4')
   cda.substanceadmin.value.addTranslationWithDisplayName(
      MMC, '628958', cda.cs.RxNorm,
      'Potassium Chloride 10 MEQ Extended Release Tablet [Klor-Con]')
   cda.substanceadmin.manufactured.setName(MM, 'Klor-Con')
   
   local PDO = cda.substanceadmin.addRelatedObservation(SA, 'SUBJ', 'OBS', 'EVN')
   cda.util.addTemplateIds(PDO,
      {{root='2.16.840.1.113883.3.88.11.83.8.1', assigningAuthorityName='HITSP C83'}})
   cda.observation.addCode(PDO, '73639000', cda.cs.SNOMED_CT, 'Prescription Drug')
   
   local SO = cda.substanceadmin.addRelatedObservation(SA, 'REFR', 'OBS', 'EVN')
   cda.util.addTemplateIds(SO, {{root='2.16.840.1.113883.10.20.1.47'}})
   cda.observation.addCode(SO, '33999-4', cda.cs.LOINC, 'Status')
   local V = cda.observation.addValueWithCode(
      SO, '421139008', cda.cs.SNOMED_CT, 'No Longer Active', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#MEDSTATUS_4')
end

local RESULTS_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.3.88.11.83.122', assigningAuthorityName='HITSP/C83'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.3.28', assigningAuthorityName='IHE PCC'}
}

local RESULTS_TABLE = [[
<table border="2" width="100%">
   <thead>
      <tr>
         <th>LOINC Code</th>
         <th>Display Name</th>
         <th>Value/Unit</th>
         <th>Abnormal Flag</th>
         <th>Date Performed</th>
      </tr>
   </thead>
   <tbody>
   </tbody>
</table>
]]

local RESULTS_PROCEDURE_TEMPLATE_IDS = {
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.19'},
   {root='2.16.840.1.113883.10.20.1.29'},
   {root='2.16.840.1.113883.3.88.11.83.17'}
}

function body.results.new(Body)
   local S = cda.section.new(Body, '30954-2', 'Diagnostic Results',
      RESULTS_TEMPLATE_IDS, 'Results')
   
   local T = cda.section.addText(S)
   T:setInner(RESULTS_TABLE)
   
   local P = cda.section.addProcedureEntry(S, 'PROC', 'EVN')
   cda.util.addTemplateIds(P, RESULTS_PROCEDURE_TEMPLATE_IDS)
   local I = cda.entry.setIdRoot(P)
   I:setAttr('nullFlavor', 'NI')
   local C = cda.entry.addCode(P)
   C:setAttr('nullFlavor', 'NI')
   cda.entry.code.addOriginalTextReference(C, '#rpNone')
   cda.entry.addTextReference(P, '#lab1')
   cda.entry.setStatusCode(P, 'active')
   local ET = cda.entry.setEffectiveTime(P)
   cda.util.date.setRange(ET, {nullFlavor='UNK'})
   
   return S, T.table.tbody
end

local RESULTS_OBSERVATION_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.31'},
   {root='2.16.840.1.113883.3.88.11.83.15.1', assigningAuthorityName='HITSP C83'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.4.13', assigningAuthorityName='IHE PCC'}
}

local function addResultsObservation(S, T, Data)
   local AbnormalMap = {H='above high normal', N='normal', L='below normal'}
   
   local TR = T:append(xml.ELEMENT, 'tr')
   TR:setAttr('ID', Data.reference)
   TR:append(xml.ELEMENT, 'td'):setText(Data.code)
   TR:append(xml.ELEMENT, 'td'):setText(Data.displayName)
   TR:append(xml.ELEMENT, 'td'):setText(Data.value..' '..Data.unit)
   TR:append(xml.ELEMENT, 'td'):setText(AbnormalMap[Data.interpretationCode])
   TR:append(xml.ELEMENT, 'td'):setText(Data.effectiveTime)
   
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   cda.util.addTemplateIds(O, RESULTS_OBSERVATION_TEMPLATE_IDS)
   cda.observation.setIdRoot(O, Data.id)
   cda.observation.addCode(O, Data.code, cda.cs.LOINC, Data.displayName)
   cda.observation.addTextReference(O, '#'..Data.reference)
   cda.observation.setStatusCode(O, Data.statusCode)
   cda.observation.setEffectiveTime(O, Data.effectiveTime, 3)
   cda.observation.addValue(O, Data.value, 'PQ', Data.unit)
   cda.observation.addInterpretationCode(O, Data.interpretationCode)
end

-- Similar to what you might get back from a database query.
-- Each item in RESULTS_OBSERVATIONS is like a row in a result set.
local RESULTS_OBSERVATIONS = {
   {id='02001c1e-a452-4459-9744-e143b27c7884', code='2951-2', displayName='Sodium',
      reference='result-summary_1', statusCode='completed', effectiveTime='02/24/2010',
      value='136', unit='mEq/L', interpretationCode='H'},
   {id='b7e32cc1-a7e2-4e94-a28c-13c0bc29f011', code='2823-3', displayName='Potassium',
      reference='result-summary_2', statusCode='completed', effectiveTime='02/24/2010',
      value='4.2', unit='mEq/L', interpretationCode='N'},
   {id='cce26bbc-298a-450d-a362-2038ce2bfe6c', code='2075-0', displayName='Chloride',
      reference='result-summary_3', statusCode='completed', effectiveTime='02/24/2010',
      value='98', unit='mEq/L', interpretationCode='N'},
   {id='38d95df5-9616-4894-948b-b80578163d99', code='2951-2', displayName='Sodium',
      reference='result-summary_4', statusCode='completed', effectiveTime='02/22/2010',
      value='128', unit='mEq/L', interpretationCode='L'},
   {id='924177d4-38d5-47f1-a54c-bef18f99baef', code='2823-3', displayName='Potassium',
      reference='result-summary_5', statusCode='completed', effectiveTime='02/22/2010',
      value='3.2', unit='mEq/L', interpretationCode='L'},
   {id='09e23418-9ae2-4568-b9cd-597f464ecbea', code='2075-0', displayName='Chloride',
      reference='result-summary_6', statusCode='completed', effectiveTime='02/22/2010',
      value='96', unit='mEq/L', interpretationCode='L'}
}

function body.results.observations(S, T)
   for i,V in ipairs(RESULTS_OBSERVATIONS) do
      addResultsObservation(S, T, V)
   end
end

local PROCEDURES_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.12'},
   {root='1.3.6.1.4.1.19376.1.5.3.1.3.11'}
}

local PROCEDURES_TABLE = [[
<table border="1" width="100%">
   <thead>
      <tr>
         <th>ICD-9-CM Code</th>
         <th>Procedure</th>
         <th>Status</th>
         <th>Date Performed</th>
      </tr>
   </thead>
   <tbody>
   </tbody>
</table>
]]

function body.procedures.new(Body)
   local S = cda.section.new(Body, '47519-4', 'Procedures',
      PROCEDURES_TEMPLATE_IDS, 'HISTORY OF PROCEDURES')
   cda.section.setId(S, 'dbec4f46-02b9-47c9-95f8-2f3688427260')
   
   local T = cda.section.addText(S)
   T:setInner(PROCEDURES_TABLE)
   
   return S, T.table.tbody
end

local PROCEDURES_TABLE_ROW_TEMPLATE = [[
<td>#</td>
<td><content ID="#">#</content></td>
<td ID="#">#</td>
<td>#</td>
]]

local function addProcedureSummary(T, Index, Code, Procedure, Status, Date)
   local TR = T:append(xml.ELEMENT, 'tr')
   TR:setInner(PROCEDURES_TABLE_ROW_TEMPLATE)
   TR[1][1] = Code
   TR[2].content.ID = 'Proc'..Index
   TR[2].content[2] = Procedure
   TR[3].ID = 'Proc'..Index..'_status'
   TR[3][2] = Status
   TR[4][1] = Date
end

local PROCEDURE_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.29'}
}

local PROCEDURE_OBSERVATION_TEMPLATE_IDS = {
   {root='2.16.840.1.113883.10.20.1.28'}
}

function body.procedures.angioplasty(S, T)
   addProcedureSummary(T, 1, '00.66',
      'Percutaneous transluminal coronary angioplasty',
      'Completed', '09/17/2007')
   
   local P = cda.section.addProcedureEntry(S, 'PROC', 'EVN', 'DRIV')
   cda.util.addTemplateIds(P, PROCEDURE_TEMPLATE_IDS)
   cda.entry.setIdRoot(P, '725731f0-bb8c-416b-86b9-384f867b9df5')
   local C = cda.entry.addCode(P, '00.66',
      cda.cs.ICD9CM_Procedure, 'Percutaneous transluminal coronary angioplasty')
   cda.entry.code.addOriginalTextReference(C, '#Proc1')
   cda.entry.setStatusCode(P, 'completed')
   cda.entry.setEffectiveTime(P, '2007-09-17', 3)
   
   local O = cda.entry.addRelatedObservation(P, 'RSON', 'OBS', 'EVN')
   cda.util.addTemplateIds(O, PROCEDURE_OBSERVATION_TEMPLATE_IDS)
   cda.observation.addCode(O, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(O, 'completed')
   local V = cda.observation.addValueWithCode(
      O, 'completed', cda.cs.ActStatus, 'Completed', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#Proc1_status')
end

function body.procedures.catheterization(S, T)
   addProcedureSummary(T, 2, '37.21',
      'Cardiac catheterization',
      'Completed', '10/01/2006')
   
   local P = cda.section.addProcedureEntry(S, 'PROC', 'EVN', 'DRIV')
   cda.util.addTemplateIds(P, PROCEDURE_TEMPLATE_IDS)
   cda.entry.setIdRoot(P, '05378884-8789-4bc0-aaa9-7d6ada25a820')
   local C = cda.entry.addCode(P, '37.21',
      cda.cs.ICD9CM_Procedure, 'Cardiac catheterization')
   cda.entry.code.addOriginalTextReference(C, '#Proc2')
   cda.entry.setStatusCode(P, 'completed')
   cda.entry.setEffectiveTime(P, '2006-10-01', 3)
   
   local O = cda.entry.addRelatedObservation(P, 'RSON', 'OBS', 'EVN')
   cda.util.addTemplateIds(O, PROCEDURE_OBSERVATION_TEMPLATE_IDS)
   cda.observation.addCode(O, '33999-4', cda.cs.LOINC, 'Status')
   cda.observation.setStatusCode(O, 'completed')
   local V = cda.observation.addValueWithCode(
      O, 'completed', cda.cs.ActStatus, 'Completed', 'CE')
   cda.observation.code.addOriginalTextReference(V, '#Proc2_status')
end

function writeToFile(FilePath, FileContent)
   local CdaOut = io.open(FilePath, 'w')
   CdaOut:write(FileContent)
   CdaOut:close()
end
