cda.entry = {}

cda.entry.code = {}
cda.entry.value = {}

-- EntryTypeCode is optional
function cda.entry.new(S, ElementName, ClassCode, MoodCode, EntryTypeCode)
   local E = S:append(xml.ELEMENT, 'entry')
   if EntryTypeCode then
      E:setAttr('typeCode', EntryTypeCode)
   end
   E:setInner('<'..ElementName..' classCode="#" moodCode="#"/>')
   local El = E[ElementName]
   El.classCode = ClassCode
   El.moodCode = MoodCode
   return El
end


--------------------------------------------------------------------------------------------------------
-- The following functions in cda.entry accept an <act>, <encounter>,                                 --
-- <observation>, <procedure>, or <substanceAdministration> element.                                  --
-- More specific functions for <observation> elements can be found in cda.observation.                --
-- More specific functions for <substanceAdministration> elements can be found in cda.substanceadmin. --
--------------------------------------------------------------------------------------------------------

-- DisplayName is optional
function cda.entry.addCode(E, Code, CodeSystem, DisplayName)
   return cda.util.addCodeElement(E, 'code', Code, CodeSystem, DisplayName)
end

-- DisplayName is optional
function cda.entry.addTargetSiteCode(E, Code, CodeSystem, DisplayName)
   return cda.util.addCodeElement(E, 'targetSiteCode', Code, CodeSystem, DisplayName)
end

function cda.entry.setStatusCode(E, StatusCode)
   local SC = imp.getOrCreateElement(E, 'statusCode')
   SC:setAttr('code', StatusCode)
   return SC
end

function cda.entry.setIdRoot(E, IdRoot)
   local I = imp.getOrCreateElement(E, 'id')
   I:setAttr('root', IdRoot)
   return I
end

-- Add a simple <code nullFlavor='?'/> element.
function cda.entry.addNullFlavor(A, NullFlavor)
   local C = A:append(xml.ELEMENT, 'code')
   return cda.entry.code.addNullFlavor(C, NullFlavor)
end

-- Returns the <text> element.  The inner text will be set to InnerText,
-- if specified.
function cda.entry.addText(S, InnerText)
   local T = S:append(xml.ELEMENT, 'text')
   if InnerText then
      S:child('text'):append(xml.TEXT, InnerText)
   end
   return T
end

-- Add a text reference, like so:
-- <text>
--    <reference value='ReferenceValue'/>
-- </text>
function cda.entry.addTextReference(E, ReferenceValue)
   local T = cda.entry.addText(E)
   local R = T:append(xml.ELEMENT, 'reference')
   R:setAttr('value', ReferenceValue)
   return T
end

-- Default precision is year only
function cda.entry.setEffectiveTime(O, FuzzyDate, Precision)
   Precision = Precision or 1
   local ET = imp.getOrCreateElement(O, 'effectiveTime')
   if FuzzyDate then
      return cda.util.date.setDate(ET, FuzzyDate, Precision)
   else
      return ET
   end
end

-- XsiType and Unit are optional.
function cda.entry.addValue(O, Value, XsiType, Unit)
   local V = O:append(xml.ELEMENT, 'value')
   if XsiType then
      cda.util.setXsiType(V, XsiType)
   end
   V:setAttr('value', Value)
   if Unit then
      V:setAttr('unit', Unit)
   end
   return V
end

function cda.entry.addRelatedAct(O, TypeCode, ClassCode, MoodCode)
   local ER = O:append(xml.ELEMENT, 'entryRelationship')
   ER:setAttr('typeCode', TypeCode)
   local A = ER:append(xml.ELEMENT, 'act')
   A:setAttr('classCode', ClassCode)
   A:setAttr('moodCode', MoodCode)
   return A
end

-- Returns an inner observation element
-- Any extra attributes for the encompassing <entryRelationship>
-- element can be specified in RelationshipAttributes.
-- For example:
-- cda.entry.addRelatedObservation(O,
--    'SUBJ', 'OBS', 'EVN', {inversionInd=false})
function cda.entry.addRelatedObservation
(
   E, TypeCode, ClassCode, MoodCode, RelationshipAttributes
)
   local ER = E:append(xml.ELEMENT, 'entryRelationship')
   ER:setAttr('typeCode', TypeCode)
   local MO = ER:append(xml.ELEMENT, 'observation')
   MO:setAttr('classCode', ClassCode)
   MO:setAttr('moodCode', MoodCode)
   if RelationshipAttributes then
      for K,V in pairs(RelationshipAttributes) do
         ER:setAttr(K, tostring(V))
      end
   end
   return MO
end

-- Extension is optional.
-- Creates a <performer> element with an inner <assignedEntity>
-- element, and returns the inner element.
function cda.entry.addPerformer(E, TypeCode, Root, Extension)
   local P = E:append(xml.ELEMENT, 'performer')
   P:setAttr('typeCode', TypeCode)
   local AE = imp.createChildWithId(P, 'assignedEntity', Root, Extension)
   return AE
end
   
   
---------------------------------------------------------------------------------------------------
-- Functions in cda.entry.code accept an element returned from addCode() or other code elements  --
---------------------------------------------------------------------------------------------------

function cda.entry.code.addOriginalText(C, OriginalText)
   local OT = C:append(xml.ELEMENT, 'originalText')
   OT:setText(OriginalText)
   return OT
end

function cda.entry.code.addOriginalTextReference(C, ReferenceVal)
   local OT = C:append(xml.ELEMENT, 'originalText')
   OT:setInner('<reference value="#"/>')
   OT.reference:setAttr('value', ReferenceVal)
   return OT
end

-- Name and Value should be tables with keys
-- code, codeSystem, and displayName
function cda.entry.code.addQualifier(C, Name, Value)
   if not Name.codeSystem then
      error('Name table must have codeSystem entry', 2)
   elseif not Value.codeSystem then
      error('Value table must have codeSystem entry', 2)
   end
   local Q = C:append(xml.ELEMENT, 'qualifier')
   cda.util.addCodeElement(Q, 'name', Name.code, Name.codeSystem, Name.displayName)
   cda.util.addCodeElement(Q, 'value', Value.code, Value.codeSystem, Value.displayName)
   return Q
end

local REGION_OF_INTEREST_TEMPLATE =
[[<regionOfInterest classCode="#" moodCode="#" ID="#">
	<id root="#"/>
	<code code="#"/>
</regionOfInterest>]]

function cda.entry.addRegionOfInterest
(
   O, TypeCode, ClassCode, MoodCode, Id, IdRoot, Code
)
   local ER = O:append(xml.ELEMENT, 'entryRelationship')
   ER:setAttr('typeCode', TypeCode)
   ER:setInner(REGION_OF_INTEREST_TEMPLATE)
   local RI = ER.regionOfInterest
   RI.classCode = ClassCode
   RI.moodCode = MoodCode
   RI.ID = Id
   RI.id.root = IdRoot
   RI.code.code = Code
   return RI
end

local OBSERVATION_MEDIA_TEMPLATE =
[[<observationMedia classCode="#" moodCode="#">
	<id root="#"/>
	<value mediaType="#">
		<reference value="#"/>
	</value>
</observationMedia>]]

function cda.entry.addObservationMedia
(
   Node, TypeCode, ClassCode, MoodCode, IdRoot, MediaType, ReferenceValue
)
   local ER = Node:append(xml.ELEMENT, 'entryRelationship')
   ER:setAttr('typeCode', TypeCode)
   ER:setInner(OBSERVATION_MEDIA_TEMPLATE)
   local OM = ER.observationMedia
   OM.classCode = ClassCode
   OM.moodCode = MoodCode
   OM.id.root = IdRoot
   OM.value.mediaType = MediaType
   OM.value.reference.value = ReferenceValue
   return OM
end

function cda.entry.code.addNullFlavor(C, NullFlavor)
   C:setAttr('nullFlavor', NullFlavor)
   return C
end

---------------------------------------------------------------------------------------------------------
-- Functions in cda.entry.value accept an element returned from addValue() or any other value elements --
---------------------------------------------------------------------------------------------------------

function cda.entry.value.addTranslationWithValue(V, Value, Code, CodeSystem)
   local T = V:append(xml.ELEMENT, 'translation')
   T:setAttr('value', Value)
   T:setAttr('code', Code)
   cda.codesystem.set(T, CodeSystem)
   return T
end

function cda.entry.value.addTranslationWithDisplayName(V, Code, CodeSystem, DisplayName)
   local T = V:append(xml.ELEMENT, 'translation')
   T:setAttr('code', Code)
   cda.codesystem.set(T, CodeSystem)
   T:setAttr('displayName', DisplayName)
   return T
end
