cda.header = {}

cda.header.author = {}
cda.header.custodian = {}
cda.header.authenticator = {}
cda.header.encounter = {}
cda.header.assigned = {}
cda.header.associated = {}


-- Use cda.header.patient functions with the return value.
-- Extension and AssigningAuthorityName are optional
function cda.header.addPatientRole(Doc, Root, Extension, AssigningAuthorityName)
   local RT = Doc.ClinicalDocument:append(xml.ELEMENT, 'recordTarget')
   local PR = imp.createChildWithId(RT, 'patientRole', Root, Extension, AssigningAuthorityName)
   return PR
end

-- Use cda.header.author functions with the return value
-- Extension is optional
function cda.header.addAuthor(Doc)
   return Doc.ClinicalDocument:append(xml.ELEMENT, 'author')
end

-- Use cda.header.custodian functions with the return value
function cda.header.addCustodian(Doc)
   return Doc.ClinicalDocument:append(xml.ELEMENT, 'custodian')
end

-- Use cda.header.authenticator functions with the return value
-- Extension is optional
function cda.header.addAuthenticator(Doc)
   return Doc.ClinicalDocument:append(xml.ELEMENT, 'legalAuthenticator')
end

local PARTICIPANT_TEMPLATE_ID_TEMPLATE = [[
<templateId root="2.16.840.1.113883.3.88.11.83.3" assigningAuthorityName="HITSP/C83"/>
<templateId root="1.3.6.1.4.1.19376.1.5.3.1.2.4" assigningAuthorityName="IHE/PCC"/>
]]

-- Use cda.participant functions with the return value
-- Time is optional
function cda.header.addParticipant(Doc, TypeCode)
   local P = Doc.ClinicalDocument:append(xml.ELEMENT, 'participant')
   P:setAttr('typeCode', TypeCode)
   P:setInner(PARTICIPANT_TEMPLATE_ID_TEMPLATE)
   return P
end

-- Use cda.document functions with the return value
function cda.header.addParentDocument(Doc, TypeCode)
   local RD = Doc.ClinicalDocument:append(xml.ELEMENT, 'relatedDocument')
   RD:setAttr('typeCode', TypeCode)
   RD:append(xml.ELEMENT, 'parentDocument')
   return RD.parentDocument
end

-- Use cda.header.encounter functions with the return value
function cda.header.addEncompassingEncounter(Doc, Root, Extension)
   local CO = Doc.ClinicalDocument:append(xml.ELEMENT, 'componentOf')
   return imp.createChildWithId(CO, 'encompassingEncounter', Root, Extension)
end

-- Use cda.header.serviceevent functions with the return value
function cda.header.addServiceEvent(Doc, TypeCode, ClassCode)
   local DOf = Doc.ClinicalDocument:append(xml.ELEMENT, 'documentationOf')
   DOf:setAttr('typeCode', TypeCode)
   local SE = DOf:append(xml.ELEMENT, 'serviceEvent')
   SE:setAttr('classCode', ClassCode)
   return SE
end

local ADD_ADDR_USAGE = [[
Add and populate an address in a CDA document 

Returns: The populated addr element.
Accepts:
 1) The parent element to which the addr element
    should be added.
 2) A table with the following optional entries
   'street'     - Street address (including number).  Can be a string, or a list of strings.
   'city'       - City
   'state'      - State
   'postalCode' - Postal (zip) code
   'country'    - Country
 3) The value of the "use" attribute (optional).

For example:

cda.util.addAddr(E, {street='353 Wine Street', city='Flint', state='MI', postalCode='48503', country='USA'}, 'HP')  
]]

function cda.header.addAddr(E, Arg, Use)
   imp.checkParam(Arg, {street=1, city=1, state=1, postalCode=1, country=1}, ADD_ADDR_USAGE)
   local A = E:append(xml.ELEMENT, 'addr')
   A:setAttr('use', Use)
   if type(Arg.street) == 'string' then
      A:append(xml.ELEMENT, 'streetAddressLine'):setText(Arg.street)
   elseif type(Arg.street) == 'table' then
      for i,V in ipairs(Arg.street) do
         A:append(xml.ELEMENT, 'streetAddressLine'):setText(V)
      end
   end
   if Arg.city       then A:append(xml.ELEMENT, 'city')             :setText(Arg.city)       end
   if Arg.state      then A:append(xml.ELEMENT, 'state')            :setText(Arg.state)      end
   if Arg.postalCode then A:append(xml.ELEMENT, 'postalCode')       :setText(Arg.postalCode) end
   if Arg.country    then A:append(xml.ELEMENT, 'country')          :setText(Arg.country)    end
   return A
end

-- Value and Use are optional
function cda.header.addTelecom(E, Value, Use)
   local T = E:append(xml.ELEMENT, 'telecom')
   if Use then
      T:setAttr('use', Use)
   end
   if Value then
      T:setAttr('value', Value)
   end
   return T
end


--------------------------------------------------------------------
-- All functions in cda.header.author accept an <author> element. --
--------------------------------------------------------------------

-- Default precision is 3 (YYYYMMDD)
function cda.header.author.setTime(A, FuzzyTime, Precision)
   local T = A.time or A:insert(1, xml.ELEMENT, 'time')
   return cda.util.date.setDate(T, FuzzyTime, Precision)
end

function cda.header.author.addAssignedAuthor(A, Root, Extension)
   return imp.createChildWithId(A, 'assignedAuthor', Root, Extension)
end


-------------------------------------------------------------------------
-- All functions in cda.header.custodian accept a <custodian> element. --
-------------------------------------------------------------------------

-- Extension is optional
-- Returns the assignedCustodian element and the inner
-- representedCustodianOrganization element
function cda.header.custodian.addAssignedCustodian(C, Name, Root, Extension)
   local AC = C:append(xml.ELEMENT, 'assignedCustodian')
   local RCO = imp.createChildWithId(AC, 'representedCustodianOrganization', Root, Extension)
   local N = RCO:append(xml.ELEMENT, 'name')
   N:setText(Name)
   return AC, RCO
end


--------------------------------------------------------------------------------------
-- All functions in cda.header.authenticator accept a <legalAuthenticator> element. --
--------------------------------------------------------------------------------------

-- Default precision is 3 (YYYYMMDD)
function cda.header.authenticator.setTime(LA, FuzzyTime, Precision)
   local T = LA.time or LA:insert(1, xml.ELEMENT, 'time')
   return cda.util.date.setDate(T, FuzzyTime, Precision)
end

function cda.header.authenticator.setSignatureCode(LA, Code)
   local SC = imp.getOrCreateElement(LA, 'signatureCode')
   SC:setAttr('code', Code)
   return SC
end

function cda.header.authenticator.addAssignedEntity(LA, Root, Extension)
   return imp.createChildWithId(LA, 'assignedEntity', Root, Extension)
end

--------------------------------------------------------------------------------------
-- All functions in cda.header.encounter accept an <encompassingEncounter> element. --
--------------------------------------------------------------------------------------

-- Default precision is 3 (YYYYMMDD)
function cda.header.encounter.setEffectiveTime(EE, FuzzyTime, Precision)
   local T = imp.getOrCreateElement(EE, 'effectiveTime')
   return cda.util.date.setDate(T, FuzzyTime, Precision)
end

function cda.header.encounter.setTypeCode(EE, Code)
   EE.encounterParticipant:setAttr('typeCode', Code)
   return EE.encounterParticipant
end

function cda.header.encounter.setParticipantTime(EE, FuzzyTime)
   return cda.util.date.setDate(EE.encounterParticipant.time, FuzzyTime)
end

function cda.header.encounter.setParticipantName(EE, Arg)
   return cda.util.name.set(EE.encounterParticipant.assignedEntity.assignedPerson.name, Arg)
end

function cda.header.encounter.setLocationCode(EE, Code)
   local C = EE.location.healthCareFacility.code
   C:setAttr('code', Code)
   return Code
end

function cda.header.encounter.setLocationDisplayName(EE, Name)
   local C = EE.location.healthCareFacility.code
   C:setAttr('displayName', Name)
   return C
end

-- Use cda.header.encounter.participant functions with the return value
function cda.header.encounter.addParticipant(EE, TypeCode)
   local EP = EE:append(xml.ELEMENT, 'encounterParticipant')
   EP:setAttr('typeCode', TypeCode)
   return EP
end

local HEALTH_CARE_FACILITY_TEMPLATE = [[
<healthCareFacility classCode="#">
	<code code="#" displayName="#"/>
</healthCareFacility>
]]

function cda.header.encounter.setHealthCareFacility(EE, ClassCode, Code, DisplayName)
   local L = imp.getOrCreateElement(EE, 'location')
   L:setInner(HEALTH_CARE_FACILITY_TEMPLATE)
   local HCF = L.healthCareFacility
   HCF.classCode = ClassCode
   HCF.code.code = Code
   HCF.code.displayName = DisplayName
   cda.codesystem.set(HCF.code, cda.cs.ServiceDeliveryLocation)
   return HCF
end


--------------------------------------------------------------------------------------------------
-- All functions in cda.header.assigned accept an <assignedAuthor> or <assignedEntity> element. --
--------------------------------------------------------------------------------------------------

-- Arg can be a standard "name table", or it can be a
-- single string.
function cda.header.assigned.addName(A, Arg)
   local AP = A:append(xml.ELEMENT, 'assignedPerson')
   return cda.util.addName(AP, Arg)
end

-- Root and Extension are optional
function cda.header.assigned.addRepresentedOrganization(A, Root, Extension, Name)
   local RO = imp.createChildWithId(A, 'representedOrganization', Root, Extension)
   if Name then
      RO:append(xml.ELEMENT, 'name'):setText(Name)
   end
   return RO
end

function cda.header.assigned.addId(A, Root, Extension, AssigningAuthorityName)
   local I = A:append(xml.ELEMENT, 'id')
   imp.setIdAttributes(I, Root, Extension, AssigningAuthorityName)
end

function cda.header.assigned.addProviderCode(A, Code, DisplayName)
   local C = A:append(xml.ELEMENT, 'code')
   C:setAttr('code', Code)
   C:setAttr('displayName', DisplayName)
   cda.codesystem.set(C, cda.cs.ProviderCodes)
   return C
end


----------------------------------------------------------------------------------
-- All functions in cda.header.associated accept an <associatedEntity> element. --
----------------------------------------------------------------------------------

function cda.header.associated.addName(A, Name)
   local AP = A:append(xml.ELEMENT, 'associatedPerson')
   local N = AP:append(xml.ELEMENT, 'name')
   N:setText(Name)
   return N
end
